from typing import Annotated
from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_camel


class CommonQualities(BaseModel):
    model_config = ConfigDict(
        extra="allow", alias_generator=to_camel, populate_by_name=True
    )

    label: Annotated[str | None, Field(validation_alias="title")] = None
    description: str | None = None
    ref: Annotated[str | None, Field(alias="sdfRef", validation_alias="$ref")] = None



from pathlib import Path
import pytest

from onedm import sdf


@pytest.fixture(scope="session")
def test_model():
    loader = sdf.SDFLoader()
    loader.load_file(Path(__file__).parent / "test.sdf.json")
    return loader.to_sdf()



"""Semantic Definition Format (SDF) for Data and Interactions of Things

https://ietf-wg-asdf.github.io/SDF/sdf.html
"""

from .data import (
    AnyData,
    ArrayData,
    BooleanData,
    Data,
    IntegerData,
    NumberData,
    ObjectData,
    StringData,
)
from .definitions import (
    Action,
    AnyProperty,
    ArrayProperty,
    BooleanProperty,
    Event,
    IntegerProperty,
    NumberProperty,
    Object,
    ObjectProperty,
    Property,
    StringProperty,
    Thing,
)
from .document import Document, Information
from .loader import SDFLoader

__all__ = [
    "SDFLoader",
    "Document",
    "Thing",
    "Object",
    "Property",
    "Action",
    "Event",
    "Data",
    "NumberProperty",
    "IntegerProperty",
    "StringProperty",
    "BooleanProperty",
    "ObjectProperty",
    "ArrayProperty",
    "AnyProperty",
    "NumberData",
    "IntegerData",
    "BooleanData",
    "StringData",
    "ObjectData",
    "ArrayData",
    "AnyData",
    "Information",
]



from onedm import sdf


def test_reference(test_model: sdf.Document):
    assert isinstance(test_model.data["Reference"], sdf.IntegerData)
    assert test_model.data["Reference"].const == 0
    assert test_model.data["Reference"].maximum == 3



from dataclasses import dataclass
import enum
from pydantic import Field, BaseModel
from typing import Annotated, Literal

import pytest

from onedm import sdf
from onedm.sdf.from_type import data_from_type


def test_integer():
    data = data_from_type(int)

    assert isinstance(data, sdf.IntegerData)
    assert not data.nullable


def test_float():
    data = data_from_type(float)

    assert isinstance(data, sdf.NumberData)
    assert not data.nullable


def test_bool():
    data = data_from_type(bool)

    assert isinstance(data, sdf.BooleanData)
    assert not data.nullable


def test_str():
    data = data_from_type(str)

    assert isinstance(data, sdf.StringData)
    assert not data.nullable


def test_bytes():
    data = data_from_type(bytes)

    assert isinstance(data, sdf.StringData)
    assert data.sdf_type == "byte-string"
    assert not data.nullable


def test_enum():
    class MyEnum(enum.Enum):
        ONE = 1
        TWO = "two"

    data = data_from_type(MyEnum)

    # Enum should be referenced
    assert data.ref == "#/sdfData/MyEnum"
    assert "MyEnum" in data.definitions
    my_enum = data.definitions["MyEnum"]

    assert isinstance(my_enum, sdf.AnyData)
    assert my_enum.choices["ONE"].const == 1
    assert my_enum.choices["TWO"].const == "two"
    assert not data.nullable


def test_int_enum():
    class MyEnum(enum.IntEnum):
        ONE = 1
        TWO = 2

    data = data_from_type(MyEnum)

    # Model should be referenced
    assert data.ref == "#/sdfData/MyEnum"
    assert "MyEnum" in data.definitions
    my_enum = data.definitions["MyEnum"]

    assert isinstance(my_enum, sdf.IntegerData)
    assert my_enum.choices["ONE"].const == 1
    assert my_enum.choices["TWO"].const == 2
    assert not data.nullable


def test_str_enum():
    class MyEnum(str, enum.Enum):
        ONE = "one"
        TWO = "two"

    data = data_from_type(MyEnum)

    # Model should be referenced
    assert data.ref == "#/sdfData/MyEnum"
    assert "MyEnum" in data.definitions
    my_enum = data.definitions["MyEnum"]

    assert isinstance(my_enum, sdf.StringData)
    assert my_enum.choices["ONE"].const == "one"
    assert my_enum.choices["TWO"].const == "two"
    assert not data.nullable


def test_union():
    data = data_from_type(int | str)

    assert len(data.choices) == 2
    assert "choice-1" in data.choices
    assert "choice-2" in data.choices
    assert data.choices["choice-1"].type == "integer"
    assert data.choices["choice-2"].type == "string"


def test_const():
    data = data_from_type(Literal["const"])

    assert data.const == "const"


def test_string_literals():
    data = data_from_type(Literal["one", "two"])

    assert isinstance(data, sdf.StringData)
    assert data.enum == ["one", "two"]
    assert not data.nullable


def test_nullable():
    data = data_from_type(int | None)

    assert isinstance(data, sdf.IntegerData)
    assert data.nullable


def test_list():
    data = data_from_type(list[str])

    assert isinstance(data, sdf.ArrayData)
    assert isinstance(data.items, sdf.StringData)
    assert not data.unique_items
    assert not data.nullable


def test_set():
    data = data_from_type(set[str])

    assert isinstance(data, sdf.ArrayData)
    assert isinstance(data.items, sdf.StringData)
    assert data.unique_items
    assert not data.nullable


def test_model():
    class TestModel(BaseModel):
        with_default: int = 2
        with_alias: Annotated[int, Field(alias="withAlias")] = 0
        optional: float | None = None
        required: bool | None

    data = data_from_type(TestModel)

    # Model should be referenced
    assert data.ref == "#/sdfData/TestModel"
    assert "TestModel" in data.definitions
    test_model = data.definitions["TestModel"]

    assert isinstance(test_model, sdf.ObjectData)
    assert not data.nullable
    assert test_model.required == ["required"]

    assert isinstance(test_model.properties["with_default"], sdf.IntegerData)
    assert test_model.properties["with_default"].default == 2
    assert not test_model.properties["with_default"].nullable

    assert "withAlias" in test_model.properties

    assert test_model.properties["required"].nullable
    assert test_model.properties["optional"].nullable


def test_dataclass():
    @dataclass
    class TestModel:
        with_default: int = 2

    data = data_from_type(TestModel)

    # Model should be referenced
    assert data.ref == "#/sdfData/TestModel"
    assert "TestModel" in data.definitions
    test_model = data.definitions["TestModel"]

    assert isinstance(test_model, sdf.ObjectData)
    assert not data.nullable

    assert isinstance(test_model.properties["with_default"], sdf.IntegerData)
    assert test_model.properties["with_default"].default == 2
    assert not test_model.properties["with_default"].nullable


def test_label():
    data = data_from_type(Annotated[int, Field(title="Test title")])

    assert data.label == "Test title"


def test_description():
    data = data_from_type(Annotated[int, Field(description="Description")])

    assert data.description == "Description"


def test_unit():
    data = data_from_type(Annotated[float, Field(json_schema_extra={"unit": "s"})])

    assert data.unit == "s"


def test_document():
    doc = sdf.Document()

    @dataclass
    class TestModel:
        with_default: int = 2

    data = data_from_type(TestModel)
    doc.things["thing"] = sdf.Thing()
    doc.things["thing"].objects["object"] = sdf.Object()
    doc.things["thing"].objects["object"].properties["prop"] = (
        sdf.definitions.property_from_data(data)
    )

    dump = doc.model_dump(by_alias=True)
    assert (
        dump["sdfThing"]["thing"]["sdfObject"]["object"]["sdfProperty"]["prop"][
            "sdfRef"
        ]
        == "#/sdfData/TestModel"
    )
    assert "TestModel" in dump["sdfData"]



from __future__ import annotations

from typing import Annotated, Literal, Tuple, Union

from pydantic import Field, NonNegativeInt, TypeAdapter

from .common import CommonQualities
from .data import (
    AnyData,
    ArrayData,
    BooleanData,
    Data,
    IntegerData,
    NumberData,
    ObjectData,
    StringData,
)


class PropertyBase:  # pylint: disable=too-few-public-methods
    observable: bool = True
    readable: bool = True
    writable: bool = True
    sdf_required: Tuple[Literal[True]] | None = None


class NumberProperty(NumberData, PropertyBase):
    pass


class IntegerProperty(IntegerData, PropertyBase):
    pass


class BooleanProperty(BooleanData, PropertyBase):
    pass


class StringProperty(StringData, PropertyBase):
    pass


class ArrayProperty(ArrayData, PropertyBase):
    pass


class ObjectProperty(ObjectData, PropertyBase):
    pass


class AnyProperty(AnyData, PropertyBase):
    pass


Property = Union[
    AnyProperty,
    Annotated[
        IntegerProperty
        | NumberProperty
        | BooleanProperty
        | StringProperty
        | ArrayProperty
        | ObjectProperty,
        Field(discriminator="type"),
    ],
]


PropertyAdapter: TypeAdapter[Property] = TypeAdapter(Property)


def property_from_data(data: Data) -> Property:
    return PropertyAdapter.validate_python(data.model_dump())


class Action(CommonQualities):
    input_data: Annotated[Data | None, Field(alias="sdfInputData")] = None
    output_data: Annotated[Data | None, Field(None, alias="sdfOutputData")] = None
    sdf_required: Tuple[Literal[True]] | None = None


class Event(CommonQualities):
    output_data: Annotated[Data | None, Field(None, alias="sdfOutputData")] = None
    sdf_required: Tuple[Literal[True]] | None = None


class Object(CommonQualities):
    properties: dict[str, Property] = Field(
        default_factory=dict,
        alias="sdfProperty",
        description="Elements of state within Things",
    )
    actions: dict[str, Action] = Field(
        default_factory=dict,
        alias="sdfAction",
        description="Commands and methods which are invoked",
    )
    events: dict[str, Event] = Field(
        default_factory=dict,
        alias="sdfEvent",
        description='"Happenings" associated with a Thing',
    )
    data: dict[str, Data] = Field(
        default_factory=dict,
        alias="sdfData",
        description=(
            "Common modeling patterns, data constraints, "
            "and semantic anchor concepts"
        ),
    )
    sdf_required: list[str | Literal[True]] = Field(default_factory=list)
    # If array of objects
    min_items: NonNegativeInt | None = None
    max_items: NonNegativeInt | None = None


class Thing(CommonQualities):
    things: dict[str, Thing] = Field(
        default_factory=dict,
        alias="sdfThing",
        description="Definition of models for complex devices",
    )
    objects: dict[str, Object] = Field(
        default_factory=dict,
        alias="sdfObject",
        description='Main "atom" of reusable semantics for model construction',
    )
    properties: dict[str, Property] = Field(
        default_factory=dict,
        alias="sdfProperty",
        description="Elements of state within Things",
    )
    actions: dict[str, Action] = Field(
        default_factory=dict,
        alias="sdfAction",
        description="Commands and methods which are invoked",
    )
    events: dict[str, Event] = Field(
        default_factory=dict,
        alias="sdfEvent",
        description='"Happenings" associated with a Thing',
    )
    data: dict[str, Data] = Field(
        default_factory=dict,
        alias="sdfData",
        description=(
            "Common modeling patterns, data constraints, "
            "and semantic anchor concepts"
        ),
    )
    sdf_required: list[str | Literal[True]] = Field(default_factory=list)
    # If array of things
    min_items: NonNegativeInt | None = None
    max_items: NonNegativeInt | None = None


Thing.model_rebuild()



import enum
import pytest
from onedm import sdf


def test_integer_validation():
    integer = sdf.IntegerData(maximum=2)
    assert integer.validate_input(2) == 2
    with pytest.raises(ValueError):
        integer.validate_input(1.5)
    # Out of range
    with pytest.raises(ValueError):
        integer.validate_input(3)


def test_string_to_int_conversion():
    assert sdf.IntegerData().validate_input("2") == 2


def test_int_enum():
    integer = sdf.IntegerData(
        sdfChoice={"ONE": sdf.IntegerData(const=1), "TWO": sdf.IntegerData(const=2)}
    )
    value = integer.validate_input(1)
    assert isinstance(value, enum.IntEnum)
    assert value.name == "ONE"
    assert value == 1


def test_int_enum_non_const_value():
    integer = sdf.IntegerData(
        sdfChoice={
            "ONE": sdf.IntegerData(const=1),
            "OTHER": sdf.IntegerData(minimum=2, maximum=5),
        }
    )
    value = integer.validate_input(2)
    assert isinstance(value, int)
    assert value == 2


def test_number_validation(test_model: sdf.Document):
    assert test_model.data["Number"].validate_input(0.5) == 0.5
    assert test_model.data["Number"].validate_input(1) == 1.0

    # Out of range
    with pytest.raises(ValueError):
        test_model.data["Number"].validate_input(100)
    # Invalid type (array)
    with pytest.raises(ValueError):
        test_model.data["Number"].validate_input([1.0])
    # Invalid type (string)
    with pytest.raises(ValueError):
        test_model.data["Number"].validate_input("string")
    # Invalid multiple of
    with pytest.raises(ValueError):
        test_model.data["Number"].validate_input(0.1)


def test_string_validation(test_model: sdf.Document):
    assert test_model.data["String"].validate_input("0123456789") == "0123456789"

    # Invalid length
    with pytest.raises(ValueError):
        test_model.data["Number"].validate_input("too short")
    # Invalid type (array)
    with pytest.raises(ValueError):
        test_model.data["Number"].validate_input(["0123456789"])


def test_nullable_validation():
    nullable_integer = sdf.IntegerData(nullable=True)
    assert nullable_integer.validate_input(None) == None


def test_non_nullable_validation():
    integer = sdf.IntegerData(nullable=False)
    with pytest.raises(ValueError):
        integer.validate_input(None)



from onedm import sdf


def test_required_as_list():
    test = sdf.Object.model_validate(
        {
            "sdfProperty": {
                "prop1": {},
                "prop2": {},
            },
            "sdfRequired": ["prop1"],
        }
    )
    assert "prop1" in test.sdf_required


def test_required_as_prop():
    test = sdf.Object.model_validate(
        {
            "sdfProperty": {
                "prop1": {
                    "type": "integer",
                    "sdfRequired": [True],
                },
                "prop2": {},
            },
        }
    )
    assert test.properties["prop1"].sdf_required



"""Conversion from native types to sdfData."""

from typing import Type

from pydantic import TypeAdapter
from pydantic.json_schema import GenerateJsonSchema
from pydantic_core import core_schema

from . import data

DataAdapter: TypeAdapter[data.Data] = TypeAdapter(data.Data)


def data_from_type(type_: Type) -> data.Data | None:
    """Create from a native Python or Pydantic type.

    None or null is not a supported type in SDF. In this case the return value
    will be None.
    """
    definition = definition_from_type(type_)
    if definition.get("type") == "null":
        # Not supported in SDF
        return None
    return DataAdapter.validate_python(definition)


def definition_from_type(type_: Type) -> dict:
    return TypeAdapter(type_).json_schema(
        ref_template="#/sdfData/{model}", schema_generator=GenerateSDF
    )


class GenerateSDF(GenerateJsonSchema):
    """Handles the differences between JSON schema and SDF

    Note that the Pydantic SDF models converts '$ref' and 'title' to 'sdfRef'
    and 'label'.
    """

    def generate_inner(self, schema: core_schema.CoreSchema):
        definition = super().generate_inner(schema)
        # In SDF everything is nullable by default while in JSON schema it is not
        definition.setdefault("nullable", False)
        return definition

    def nullable_schema(self, schema: core_schema.NullableSchema):
        definition = self.generate_inner(schema["schema"])
        # SDF uses the nullable attribute rather than anyOf/oneOf
        definition["nullable"] = True
        return definition

    def enum_schema(self, schema: core_schema.EnumSchema):
        definition = super().enum_schema(schema)
        if "enum" in definition:
            # Replace enum with sdfChoice
            definition["sdfChoice"] = {
                member.name: {"const": member.value} for member in schema["members"]
            }
            del definition["enum"]
        return definition

    def union_schema(self, schema: core_schema.UnionSchema):
        definition = super().union_schema(schema)
        definition["sdfChoice"] = {
            f"choice-{i}": choice
            for i, choice in enumerate(definition["anyOf"], start=1)
        }
        del definition["anyOf"]
        return definition

    def bytes_schema(self, schema: core_schema.BytesSchema):
        definition = super().bytes_schema(schema)
        definition["sdfType"] = "byte-string"
        return definition

    def timedelta_schema(self, schema: core_schema.TimedeltaSchema):
        definition = super().timedelta_schema(schema)
        if definition["type"] == "number":
            definition["unit"] = "s"
        return definition



"""Data qualities.

Contains data qualities as defined by the standard, but divided into separate
types for correct validation and type hints.
"""

from __future__ import annotations

import datetime
from abc import ABC
from enum import EnumMeta, IntEnum
from re import Pattern
from typing import Annotated, Any, Literal, Union

from pydantic import Field, NonNegativeInt, field_serializer
from pydantic_core import SchemaValidator, core_schema

from .common import CommonQualities


class DataQualities(CommonQualities, ABC):
    """Base class for all data qualities."""

    type: Literal["boolean", "number", "integer", "string", "object", "array", None] = (
        None
    )
    sdf_type: Annotated[str, Field(pattern=r"^[a-z][\-a-z0-9]*$")] | None = None
    unit: str | None = None
    nullable: bool = True
    const: Any | None = None
    default: Any | None = None
    choices: Annotated[dict[str, DataQualities] | None, Field(alias="sdfChoice")] = None

    # Temporary storage for referenced models, will not be included in final document
    definitions: dict[str, Data] = Field(default_factory=dict, alias="$defs")

    def _get_base_schema(self) -> core_schema.CoreSchema:
        """Implemented by sub-classes."""
        raise NotImplementedError

    def get_pydantic_schema(self) -> core_schema.CoreSchema:
        """Get the Pydantic schema for this data quality."""
        schema: core_schema.CoreSchema

        if self.const is not None:
            schema = core_schema.literal_schema([self.const])
        elif self.choices is not None:
            schema = core_schema.union_schema(
                [
                    (choice.get_pydantic_schema(), name)
                    for name, choice in self.choices.items()
                ]
            )
        else:
            schema = self._get_base_schema()

        if self.default is not None:
            schema = core_schema.with_default_schema(schema, default=self.default)
        if self.nullable:
            schema = core_schema.nullable_schema(schema)
        return schema

    def validate_input(self, value: Any) -> Any:
        """Validate and coerce a value."""
        return SchemaValidator(self.get_pydantic_schema()).validate_python(value)


class NumberData(DataQualities):
    type: Literal["number"] = "number"
    minimum: float | None = None
    maximum: float | None = None
    exclusive_minimum: float | None = None
    exclusive_maximum: float | None = None
    multiple_of: float | None = None
    format: str | None = None
    const: float | None = None
    default: float | None = None

    @field_serializer("type")
    def always_include_type(self, type_: str, _):
        return type_

    def _get_base_schema(self) -> core_schema.CoreSchema:
        if self.sdf_type == "unix-time":
            return core_schema.datetime_schema(
                ge=(
                    datetime.datetime.fromtimestamp(self.minimum)
                    if self.minimum is not None
                    else None
                ),
                le=(
                    datetime.datetime.fromtimestamp(self.maximum)
                    if self.maximum is not None
                    else None
                ),
                gt=(
                    datetime.datetime.fromtimestamp(self.exclusive_minimum)
                    if self.exclusive_minimum is not None
                    else None
                ),
                lt=(
                    datetime.datetime.fromtimestamp(self.exclusive_maximum)
                    if self.exclusive_maximum is not None
                    else None
                ),
            )
        return core_schema.float_schema(
            ge=self.minimum,
            le=self.maximum,
            gt=self.exclusive_minimum,
            lt=self.exclusive_maximum,
            multiple_of=self.multiple_of,
        )


class IntegerData(DataQualities):
    type: Literal["integer"] = "integer"
    minimum: int | None = None
    maximum: int | None = None
    exclusive_minimum: int | None = None
    exclusive_maximum: int | None = None
    multiple_of: int | None = None
    choices: Annotated[dict[str, IntegerData] | None, Field(alias="sdfChoice")] = (
        None  # type: ignore[assignment]
    )
    const: int | None = None
    default: int | None = None
    _enum = None

    @field_serializer("type")
    def always_include_type(self, type_: str, _):
        return type_

    def _get_base_schema(self) -> core_schema.IntSchema:
        return core_schema.int_schema(
            ge=self.minimum,
            le=self.maximum,
            gt=self.exclusive_minimum,
            lt=self.exclusive_maximum,
            multiple_of=self.multiple_of,
        )

    def to_enum(self) -> EnumMeta | None:
        """Turn sdfChoice into a Python enumeration

        Only choices with the const attribute set will be included.
        """
        if self._enum is None:
            if self.choices is None:
                return None
            self._enum = IntEnum(  # type: ignore
                self.label or "Enum",
                {
                    name: choice.const
                    for name, choice in self.choices.items()
                    if choice.const is not None
                },
            )
        return self._enum

    def validate_input(self, value: Any) -> IntEnum | int:
        value = super().validate_input(value)
        # Convert to enum.IntEnum if possible
        if enum_cls := self.to_enum():
            try:
                return enum_cls(value)
            except ValueError:
                # Value is valid but not a specific enum value
                pass
        return value


class BooleanData(DataQualities):
    type: Literal["boolean"] = "boolean"
    const: bool | None = None
    default: bool | None = None
    choices: Annotated[dict[str, BooleanData] | None, Field(alias="sdfChoice")] = (
        None  # type: ignore[assignment]
    )

    @field_serializer("type")
    def always_include_type(self, type_: str, _):
        return type_

    def _get_base_schema(self) -> core_schema.BoolSchema:
        return core_schema.bool_schema()


class StringData(DataQualities):
    type: Literal["string"] = "string"
    enum: list[str] | None = None
    min_length: NonNegativeInt = 0
    max_length: NonNegativeInt | None = None
    pattern: str | Pattern[str] | None = None
    format: str | None = None
    content_format: str | None = None
    choices: Annotated[dict[str, StringData] | None, Field(alias="sdfChoice")] = (
        None  # type: ignore[assignment]
    )
    const: str | None = None
    default: str | None = None

    @field_serializer("type")
    def always_include_type(self, type_: str, _):
        return type_

    def _get_base_schema(self) -> core_schema.CoreSchema:
        if self.enum is not None:
            return core_schema.literal_schema(self.enum)
        if self.sdf_type == "byte-string" or self.format == "bytes":
            return core_schema.bytes_schema(
                min_length=self.min_length, max_length=self.max_length
            )
        if self.format == "uuid":
            return core_schema.uuid_schema()
        if self.format == "date-time":
            return core_schema.datetime_schema()
        if self.format == "date":
            return core_schema.date_schema()
        if self.format == "time":
            return core_schema.time_schema()
        if self.format == "uri":
            return core_schema.url_schema()
        return core_schema.str_schema(
            min_length=self.min_length,
            max_length=self.max_length,
            pattern=self.pattern,
        )


class ArrayData(DataQualities):
    type: Literal["array"] = "array"
    items: Data | None = None
    min_items: NonNegativeInt = 0
    max_items: NonNegativeInt | None = None
    unique_items: bool = False
    const: list | None = None
    default: list | None = None

    @field_serializer("type")
    def always_include_type(self, type_: str, _):
        return type_

    def _get_base_schema(self) -> core_schema.ListSchema | core_schema.SetSchema:
        if self.unique_items:
            return core_schema.set_schema(
                self.items.get_pydantic_schema() if self.items is not None else None,
                min_length=self.min_items,
                max_length=self.max_items,
            )
        return core_schema.list_schema(
            self.items.get_pydantic_schema() if self.items is not None else None,
            min_length=self.min_items,
            max_length=self.max_items,
        )


class ObjectData(DataQualities):
    type: Literal["object"] = "object"
    properties: dict[str, Data] | None = None
    required: list[str] = Field(default_factory=list)
    const: dict[str, Any] | None = None
    default: dict[str, Any] | None = None

    @field_serializer("type")
    def always_include_type(self, type_: str, _):
        return type_

    def _get_base_schema(self) -> core_schema.CoreSchema:
        if self.properties is None:
            return core_schema.dict_schema()
        return core_schema.typed_dict_schema(
            {
                name: core_schema.typed_dict_field(
                    property.get_pydantic_schema(), required=name in self.required
                )
                for name, property in self.properties.items()
            }
        )


class AnyData(DataQualities):
    type: Literal[None] = None

    def _get_base_schema(self) -> core_schema.AnySchema:
        return core_schema.any_schema()


Data = Union[
    AnyData,
    Annotated[
        IntegerData | NumberData | BooleanData | StringData | ObjectData | ArrayData,
        Field(discriminator="type"),
    ],
]


ObjectData.model_rebuild()
ArrayData.model_rebuild()
DataQualities.model_rebuild()
AnyData.model_rebuild()



from onedm import sdf


def test_document_generation():
    doc = sdf.Document(
        info=sdf.Information(
            title="Test title",
        ),
        things={
            "MyThing": sdf.Thing(
                things={
                    "MySubThing": sdf.Thing(
                        objects={
                            "MyObject": sdf.Object(
                                properties={
                                    "MyProperty": sdf.ArrayProperty(
                                        items=sdf.AnyData(
                                            ref="#/sdfData/MyEnum",
                                            # Put a temporary definition here
                                        ),
                                        definitions={
                                            "MyEnum": sdf.IntegerData(
                                                choices={
                                                    "ONE": sdf.IntegerData(
                                                        const=1
                                                    )
                                                }
                                            )
                                        }
                                    )
                                }
                            )
                        }
                    )
                }
            )
        }
    )

    dump = doc.model_dump(mode="json", by_alias=True, exclude_defaults=True)

    assert dump == {
        "info": {
            "title": "Test title"
        },
        "sdfThing": {
            "MyThing": {
                "sdfThing": {
                    "MySubThing": {
                        "sdfObject": {
                            "MyObject": {
                                "sdfProperty": {
                                    "MyProperty": {
                                        "type": "array",
                                        "items": {
                                            "sdfRef": "#/sdfData/MyEnum"
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        },
        "sdfData": {
            "MyEnum": {
                "type": "integer",
                "sdfChoice": {
                    "ONE": {
                        "type": "integer",
                        "const": 1
                    }
                }
            }
        }
    }



"""Loading of SDF files

Extensions supported compared to standard:

* Relative URI references

Takes care of dereferencing.
"""

import io
import json
import urllib.parse
from typing import Any

from .document import Document


class SDFLoader:

    def __init__(self) -> None:
        self.url = ""
        self.root: dict[str, Any] = {}
        self._namespaces: dict[str, SDFLoader] = {}

    def load_file(self, path):
        self.url = str(path)
        with open(path, "r", encoding="utf-8") as fp:
            self.load_from_fp(fp)

    def load_from_fp(self, fp: io.TextIOBase):
        self.root = json.load(fp)
        self._dereference(self.root)

    def load_from_dict(self, doc: dict, url: str = ""):
        self.url = url
        self.root = doc
        self._dereference(self.root)

    def load(self, url: str):
        result = urllib.parse.urlparse(url)
        if not result.scheme:
            self.load_file(url)
        else:
            raise NotImplementedError("Not supported yet")

    def to_sdf(self) -> Document:
        return Document.model_validate(self.root)

    def _dereference(self, definition: dict[str, Any]) -> dict[str, Any]:
        while "sdfRef" in definition:
            # This reference will be used to patch the referenced original
            patch = definition.copy()
            ref: str = patch.pop("sdfRef")

            if ":" in ref:
                # Resolve namespaces
                namespace, path = ref.split(":")
                # Check if this namespace has already been loaded
                if namespace in self._namespaces:
                    ref_root = self._namespaces[namespace].root
                else:
                    # Extension to standard, support relative URIs in namespaces
                    ref_url = urllib.parse.urljoin(
                        self.url, self.root["namespace"][namespace]
                    )
                    loader = SDFLoader()
                    loader.load(ref_url)
                    # Cache the loaded namespace
                    self._namespaces[namespace] = loader

                    ref_root = loader.root
                    ref_url = urllib.parse.urljoin(ref_url, path)
            else:
                ref_url = urllib.parse.urljoin(self.url, ref)
                ref_root = self.root

            result = urllib.parse.urlparse(ref_url)

            fragments = result.fragment.split("/")
            # Traverse down the local tree
            original = ref_root
            for fragment in fragments[1:]:
                if fragment not in original:
                    raise ValueError(f"Could not find {fragment} in {result.geturl()}")
                original = original[fragment]

            if patch:
                definition = merge_patch(original, patch)
            else:
                # Nothing to patch
                definition = original

        # Continue processing children
        for name, value in definition.items():
            if isinstance(value, dict):
                definition[name] = self._dereference(value)

        return definition


def merge_patch(target: dict, patch) -> dict:
    if not isinstance(patch, dict):
        return patch
    patched = target.copy()
    for name, value in patch.items():
        if value is None and name in target:
            del patched[name]
        else:
            patched[name] = merge_patch(target.get(name, {}), value)
    return patched



from datetime import datetime
from typing import Annotated, Any

from pydantic import BaseModel, ConfigDict, Field, field_serializer, model_serializer
from pydantic.alias_generators import to_camel

from . import definitions


class Information(BaseModel):
    """Information block

    The information block contains generic metadata for the SDF document itself
    and all included definitions.
    """

    title: Annotated[
        str | None,
        Field(description="A short summary to be displayed in search results, etc."),
    ] = None
    description: Annotated[
        str | None,
        Field(description="Long-form text description (no constraints)"),
    ] = None
    version: Annotated[
        str | None,
        Field(description="The incremental version of the definition"),
    ] = None
    modified: Annotated[
        datetime | None,
        Field(description="Time of the latest modification"),
    ] = None
    copyright: Annotated[
        str | None,
        Field(
            description="Link to text or embedded text containing a copyright notice"
        ),
    ] = None
    license: Annotated[
        str | None,
        Field(description="Link to text or embedded text containing license terms"),
    ] = None
    features: list[str] = Field(
        default_factory=list, description="List of extension features used"
    )


class Document(BaseModel):
    model_config = ConfigDict(
        extra="allow", alias_generator=to_camel, populate_by_name=True
    )

    info: Information = Field(default_factory=Information)
    namespace: dict[str, str] = Field(
        default_factory=dict,
        description=(
            "Defines short names mapped to namespace URIs, "
            "to be used as identifier prefixes"
        ),
    )
    default_namespace: Annotated[
        str | None,
        Field(
            description=(
                "Identifies one of the prefixes in the namespace map "
                "to be used as a default in resolving identifiers"
            ),
        ),
    ] = None
    things: dict[str, definitions.Thing] = Field(
        default_factory=dict,
        alias="sdfThing",
        description="Definition of models for complex devices",
    )
    objects: dict[str, definitions.Object] = Field(
        default_factory=dict,
        alias="sdfObject",
        description='Main "atom" of reusable semantics for model construction',
    )
    properties: dict[str, definitions.Property] = Field(
        default_factory=dict,
        alias="sdfProperty",
        description="Elements of state within Things",
    )
    actions: dict[str, definitions.Action] = Field(
        default_factory=dict,
        alias="sdfAction",
        description="Commands and methods which are invoked",
    )
    events: dict[str, definitions.Event] = Field(
        default_factory=dict,
        alias="sdfEvent",
        description='"Happenings" associated with a Thing',
    )
    data: dict[str, definitions.Data] = Field(
        default_factory=dict,
        alias="sdfData",
        description=(
            "Common modeling patterns, data constraints, "
            "and semantic anchor concepts"
        ),
    )

    def to_json(self) -> str:
        return self.model_dump_json(indent=2, exclude_defaults=True, by_alias=True)

    @field_serializer("data", mode="wrap")
    def populate_sdf_data(self, data: dict, nxt):
        """Populate sdfData

        Scans through the whole document looking for $defs and collects them
        in the document's root #/sdfData.
        """
        data = data.copy()

        def update_from_parent(
            parent: Document | definitions.Object | definitions.Thing,
        ):
            for prop in parent.properties.values():
                data.update(prop.definitions)
            for action in parent.actions.values():
                if action.input_data:
                    data.update(action.input_data.definitions)
                if action.output_data:
                    data.update(action.output_data.definitions)
            for event in parent.events.values():
                if event.output_data:
                    data.update(event.output_data.definitions)

            if isinstance(parent, definitions.Thing):
                for thing in parent.things.values():
                    update_from_parent(thing)
                for obj in parent.objects.values():
                    update_from_parent(obj)

        update_from_parent(self)
        for thing in self.things.values():  # pylint: disable=no-member
            update_from_parent(thing)

        return nxt(data)

    @model_serializer(mode="wrap")
    def remove_definitions(self, nxt):
        """Remove $ref and $defs

        These are temporary entries created by Pydantic.
        """
        doc = nxt(self)
        remove_definitions(doc)
        return doc


def remove_definitions(obj: dict[str, Any]):
    for key, value in list(obj.items()):
        if key == "$defs":
            del obj[key]
        if isinstance(value, dict):
            remove_definitions(value)
